/****************************************************************************************************/
/** @file
*
* @par Module Name:
*      drv_can.c
*
* @par Responsible:
*      Christian Gaertner
*
* @par Module Description:
*      S12/S12X MSCAN driver \n
*      Supports up to five MSCAN modules with different configurations and different physical \n
*      transceivers
*
*****************************************************************************************************/
/*
* $Date: 2007-06-17 10:00:08 +0200 (So, 17 Jun 2007) $
* $Revision: 14 $
* $Author: Christian Gaertner $
* $HeadURL: file:///D:/SVN/Drivers/trunk/CAN/drv_can.c $
*
*****************************************************************************************************/

/*****************************************************************************************************
* Module definition for conditional compiling
*****************************************************************************************************/
#define MODULE_DRV_CAN_C

/*****************************************************************************************************
* Include files
*****************************************************************************************************/
/* core modules */
#include "stdtypes.h"
#include "gdefs.h"

/* own headers */
#include "drv_can.h"

/* used modules */

/*****************************************************************************************************
* Declaration of module wide FUNCTIONs - NOT for use in other modules
*****************************************************************************************************/
void CAN_InitSetPortPins(CAN_Init_Struct *pCAN_Channel_Init);
uint8_t CAN_ReceiveFct(CoreCANstruct *pCAN_Channel);
void CAN_GetChannelAddress(uint8_t channel, CoreCANstruct * address);

/*****************************************************************************************************
* Definition of module wide MACROs / #DEFINE-CONSTANTs - NOT for use in other modules
*****************************************************************************************************/

/*****************************************************************************************************
* Declaration of module wide TYPEs - NOT for use in other modules
*****************************************************************************************************/

/*****************************************************************************************************
* Definition of module wide VARIABLEs - NOT for use in other modules
*****************************************************************************************************/
t_copyallfct * CAN_ApplicationNotify;

/*****************************************************************************************************
* Definition of module wide (CONST-) CONSTANTs - NOT for use in other modules
*****************************************************************************************************/
const uint8_t CAN_Channels = (                              /* mark the enabled CAN channels */
    #ifdef USE_CAN0
        0x01                                                /* channel 0 enabled */
    #endif
  );

/*****************************************************************************************************
* Link information
*****************************************************************************************************/
LINK_INFO_DERIVATIVE

/****************************************************************************************************/
/**
 * @brief  call all CAN init functions
 * @par    Description:
 *         Calls the port pin init and CAN hardware module init functions of all \n
 *         enabled channels.
 * @return void
 */
void CAN_Init(void)
{
    CAN_ApplicationNotify = NULL;                           /* init pointer */

    #ifdef USE_CAN0
        CAN_InitSetPortPins(&CAN_Init_CAN0);
        CAN_InitChannel(&CoreRXCAN0, &CAN_Init_CAN0);
    #endif
}

/****************************************************************************************************/
/**
 * @brief  CAN port pin initialisation
 * @par    Description:
 *         init the CAN transceiver by setting the dedicated port pins
 * @param  *pCAN_Channel_Init  pointer to the const CAN init struct
 * @return void
 */
void CAN_InitSetPortPins(CAN_Init_Struct *pCAN_Channel_Init)
{
    switch (pCAN_Channel_Init->transceiver)
    {
        case TJA1040:
        case TJA1050:
            /* set stndby pin (output) */
            *(pCAN_Channel_Init->portpins.stndby_dir) |= pCAN_Channel_Init->portpins.stndby_mask;
            *(pCAN_Channel_Init->portpins.stndby) &= (uint8_t)(~pCAN_Channel_Init->portpins.stndby_mask);
            break;
        case TJA1041:
        case TJA1054:                                       /* switch transceiver to normal mode */
            /* set enable pin (output) */
            *(pCAN_Channel_Init->portpins.enable_dir) |= pCAN_Channel_Init->portpins.enable_mask;
            *(pCAN_Channel_Init->portpins.enable) |= pCAN_Channel_Init->portpins.enable_mask;
            /* set stndby pin (output) */
            *(pCAN_Channel_Init->portpins.stndby_dir) |= pCAN_Channel_Init->portpins.stndby_mask;
            *(pCAN_Channel_Init->portpins.stndby) |= pCAN_Channel_Init->portpins.stndby_mask;
            /* set error pin (input) */
            *(pCAN_Channel_Init->portpins.stndby) &= (uint8_t)(~pCAN_Channel_Init->portpins.stndby_mask);
            *(pCAN_Channel_Init->portpins.stndby_dir) &= (uint8_t)(~pCAN_Channel_Init->portpins.stndby_mask);
            break;
        default:
            break;
    }
}

/****************************************************************************************************/
/**
 * @brief  CAN hardware init
 * @par    Description:
 *         Inits the CAN hardware with the predefined constants.
 * @param  *pCAN_Channel       pointer to the CAN struct
 * @param  *pCAN_Channel_Init  pointer to the const CAN init struct
 * @return void
 */
void CAN_InitChannel(CoreCANstruct *pCAN_Channel, CAN_Init_Struct *pCAN_Channel_Init)
{
    uint8_t ccr;

    SAVE_CCR(ccr);                                          /* buffer state of CCR */
    DISABLE_IRQS();                                         /* prevent interrupts */

    pCAN_Channel->pCAN->canctl0.Byte = INITRQ;              /* request Init Mode */
    while(pCAN_Channel->pCAN->canctl1.Bits.initak == CLEAR)
    {                                                       /* Wait till Acknowledge */
    }
                                                            /* control register 1 */
    pCAN_Channel->pCAN->canctl1.Byte = pCAN_Channel_Init->canctl1.Byte;
                                                            /* bit timing, CAN speed */
    pCAN_Channel->pCAN->canbtr0.Byte = pCAN_Channel_Init->canbtr0.Byte;
    pCAN_Channel->pCAN->canbtr1.Byte = pCAN_Channel_Init->canbtr1.Byte;
                                                            /* filter config */
    pCAN_Channel->pCAN->canidac.Byte = pCAN_Channel_Init->canidac.Byte;
                                                            /* filter acceptance/mask registers */
    pCAN_Channel->pCAN->canid[0] = pCAN_Channel_Init->canid[0];
    pCAN_Channel->pCAN->canid[1] = pCAN_Channel_Init->canid[1];
                                                            /* clear INITRQ */
    pCAN_Channel->pCAN->canctl0.Byte = 0;

    while(pCAN_Channel->pCAN->canctl1.Bits.initak != 0)
    {                                                       /* wait for INITAK to clear */
    }
                                                            /* control register 0 */
    pCAN_Channel->pCAN->canctl0.Byte = pCAN_Channel_Init->canctl0.Byte;
                                                            /* receive interrupts */
    pCAN_Channel->pCAN->canrier.Byte = pCAN_Channel_Init->canrier.Byte;
                                                            /* transmit interrupts */
    pCAN_Channel->pCAN->cantier.Byte = pCAN_Channel_Init->cantier.Byte;

    while(pCAN_Channel->pCAN->canctl0.Bits.synch != 1)
    {                                                       /* wait for MSCAN to synchronize */
    }

    RESTORE_CCR(ccr);
}

/****************************************************************************************************/
/**
 * @brief  set notify function
 * @par    Description:
 *         Sets the address of the notify function to the function pointer to enable \n
 *         thiscommunication way
 * @param  *pFctAddress - address of the notify function
 * @return void
 */
void CAN_SetNotifyFunction(t_copyallfct * pFctAddress)
{
    CAN_ApplicationNotify = pFctAddress;
}

/****************************************************************************************************/
/**
 * @brief  Set Standby Node
 * @par    Description:
 *         Activates the standby mode of the specific CAN channel transceiver.
 * @param  *pCAN_Channel       pointer to the CAN struct
 * @return void
 */
void CAN_Set_Standby_Mode(CAN_Init_Struct *pCAN_Channel_Init)
{
    switch (pCAN_Channel_Init->transceiver)
    {
        case TJA1040:
        case TJA1050:                                       /* transceiver switches to SILENT mode */
                                                            /* set stndby pin (output) */
            *(pCAN_Channel_Init->portpins.stndby) |= pCAN_Channel_Init->portpins.stndby_mask;
            break;
        case TJA1041:
        case TJA1054:
                                                            /* set enable pin (output) */
            *(pCAN_Channel_Init->portpins.enable) &= (uint8_t)(~pCAN_Channel_Init->portpins.enable_mask);
                                                            /* set stndby pin (output) */
            *(pCAN_Channel_Init->portpins.stndby) &= (uint8_t)(~pCAN_Channel_Init->portpins.stndby_mask);
            break;
        default:
            break;
    }
}

/****************************************************************************************************/
/**
 * @brief  CAN receive function
 * @par    Description:
 *         Called by the receive interrupts
 * @param  CoreCANstruct *pCAN_Channel
 * @return void
 */
bool_t CAN_ReceiveFct(CoreCANstruct *pCAN_Channel)
{
    volatile uint8_t i;
    volatile uint32_t Received_ID = 0;
    bool_t retvalue = FALSE;

    // @todo handling of RTR

    // @todo copy received ID


    if (Received_ID & 0x0008)                               /* check for extended ID (IDE = 1) */
    {
        Received_ID = ( (pCAN_Channel->pCAN->rxbuf.id.l & 0x0007FFFF) >> 1 );  /* Bits 17 - 0, all others are set to 0 */
        Received_ID |= ( (pCAN_Channel->pCAN->rxbuf.id.l & 0xFFE00000) >> 3 ); /* Bits 28 - 18 */
        Received_ID |= CAN_MARK_EXTENDED_ID;                /* mark as internal ID */
    }
    else                                                    /* standard ID */
    {
        Received_ID = ( pCAN_Channel->pCAN->rxbuf.id.w[0] >> 5 );
    }

    for (i=0; i<CAN_MAX_MSGBUFFER; i++)
    {
        if (Received_ID == pCAN_Channel->id[i])             /* if ID matches... */
        {
            if (pCAN_Channel->precopyfct[i] != NULL)
            {                                               /* ...call copy function and leave loop */
                pCAN_Channel->precopyfct[i](pCAN_Channel->pCAN->rxbuf.dlr.Byte, &(pCAN_Channel->pCAN->rxbuf.dsr.b[0]));
                break;
            }
            else
            {                                               /* ...or store message and set flag */
                pCAN_Channel->type[i] |= MESSAGE_RECEIVED_NEW;
                pCAN_Channel->timestamp[i] = pCAN_Channel->pCAN->rxbuf.tsr;
                pCAN_Channel->len[i] = pCAN_Channel->pCAN->rxbuf.dlr.Byte;
                *((uint16_t*)&pCAN_Channel->data[i][0]) = pCAN_Channel->pCAN->rxbuf.dsr.w[0];
                *((uint16_t*)&pCAN_Channel->data[i][2]) = pCAN_Channel->pCAN->rxbuf.dsr.w[1];
                *((uint16_t*)&pCAN_Channel->data[i][4]) = pCAN_Channel->pCAN->rxbuf.dsr.w[2];
                *((uint16_t*)&pCAN_Channel->data[i][6]) = pCAN_Channel->pCAN->rxbuf.dsr.w[3];
                retvalue = TRUE;
            }
        }
        else
        {                                                   /* ID not predefined - if copy function not NULL, copy it */
            if (NULL != CAN_ApplicationNotify)
            {
                /**
                 *  @note CAN_ApplicationNotify - Function is only called if address is
                 *        set by CAN_SetNotifyFunction().
                 */
                CAN_ApplicationNotify(Received_ID, pCAN_Channel->pCAN->rxbuf.dlr.Byte, &(pCAN_Channel->pCAN->rxbuf.dsr.b[0]));
            }
        }
    }

    pCAN_Channel->pCAN->canrflg.Bits.rxf = 1;               /* clear Rx flag */

    return (retvalue);
}

/****************************************************************************************************/
/**
 * @brief  send messages
 * @par    Description:
 *         checks transmit flag for message to be send \n
 *         copies id, data, length to the transmit buffer
 * @param  CoreCANstruct *pCAN_Channel - CAN structure selects hardware
 * @return void
 */
void CAN_TransmitFct(CoreCANstruct * pCAN_Channel)
{
    uint8_t i;

    if (pCAN_Channel->pCAN->cantflg.Byte == 0)              /* check for empty transmit buffer */
    {
                                                            /* disable interrupts for empty buffers */
        pCAN_Channel->pCAN->cantier.Byte = (uint8_t)(~(pCAN_Channel->pCAN->cantflg.Byte) & (TXE0 | TXE1 | TXE2));


        for (i=0; i<CAN_MAX_MSGBUFFER; i++)                 /* search for message to send */
        {
            if ( (pCAN_Channel->type[i] & MESSAGE_READY_TO_SEND) == MESSAGE_READY_TO_SEND )
            {                                               /* message selected, delete flag */
                pCAN_Channel->type[i] &= (uint8_t)(~MESSAGE_READY_TO_SEND);
                                                            /* select Tx buffer */
                pCAN_Channel->pCAN->cantbsel.Byte = pCAN_Channel->pCAN->cantflg.Byte;
                                                            /* copy id */
                if (pCAN_Channel->id[i] & CAN_MARK_EXTENDED_ID)
                {                                           /* check for extended ID */
                    pCAN_Channel->pCAN->txbuf.id.l = ( (pCAN_Channel->id[i] << 3) & 0xFFE00000 ); /* Bits 28 - 18 */
                    pCAN_Channel->pCAN->txbuf.id.l |= ( (pCAN_Channel->id[i] & 0x0003FFFF) << 1); /* Bits 17 - 0 */
                    pCAN_Channel->pCAN->txbuf.id.w[0] |= (uint16_t)MSCAN_SRR_MASK | MSCAN_IDE_MASK; /* SRR=1, IDE=1 */
                }
                else
                {                                           /* standard id */
                    pCAN_Channel->pCAN->txbuf.id.w[0] = (uint16_t)( pCAN_Channel->id[i] << 5 );
                }

                                                            /* copy always 8 byte data from buffer */
                (pCAN_Channel->pCAN->txbuf.dsr.w[0]) = *((uint16_t*)&(pCAN_Channel->data[i][0]));
                (pCAN_Channel->pCAN->txbuf.dsr.w[1]) = *((uint16_t*)&(pCAN_Channel->data[i][2]));
                (pCAN_Channel->pCAN->txbuf.dsr.w[2]) = *((uint16_t*)&(pCAN_Channel->data[i][4]));
                (pCAN_Channel->pCAN->txbuf.dsr.w[3]) = *((uint16_t*)&(pCAN_Channel->data[i][6]));

                                                            /* copy length */
                pCAN_Channel->pCAN->txbuf.dlr.Byte = pCAN_Channel->len[i];
                                                            /* queue CAN transmit buffer */
                pCAN_Channel->pCAN->cantflg.Byte = pCAN_Channel->pCAN->cantbsel.Byte;
                                                            /* enable interrupt for this buffer */
                pCAN_Channel->pCAN->cantier.Byte |= pCAN_Channel->pCAN->cantbsel.Byte;
            }
        }
    }
    else
    {
        /**
         *  @note if no empty tx buffer is available the message is not send \n
         *        normally this should not happen, because tx interrupt is called
         *        only if tx buffer is empty (message sent). \n
         *        Possible case: interrupts are locked, tx interrupt is pending and application
         *        just filled the last buffer using the CAN_TransmitMsg() function. \n
         *        So do not disable interrupt here - after at least one buffer is empty
         *        this function is called automatically again. \n
         */
    }
}

/****************************************************************************************************/
/**
 * @brief  CAN transmit function
 * @par    Description:
 *         Copy parameters to hardware, transmit message
 * @param  CoreCANstruct *pCAN_Channel  - CAN structure selects hardware
 * @param  uint32_t  CAN_Id             - CAN identifier
 * @param  uint8_t   CAN_Datalength     - data length
 * @param  uint8_t * CAN_Data           - reference to data
 * @return TRUE - message send, FALSE - message not send
 */
bool_t CAN_TransmitMsg(CoreCANstruct *pCAN_Channel, uint32_t dw_id_l, uint8_t dlc, uint8_t *data)
{

    if (pCAN_Channel->pCAN->cantflg.Byte == 0)
    {
        return (FALSE);                                     /* no empty tx buffer, message not send */
    }
                                                            /* automatic channel select */
    pCAN_Channel->pCAN->cantbsel.Byte = pCAN_Channel->pCAN->cantflg.Byte;
                                                            /* send extended Id */
    if ((dw_id_l & CAN_MARK_EXTENDED_ID) == CAN_MARK_EXTENDED_ID)
    {
        pCAN_Channel->pCAN->txbuf.id.b[3] = (uint8_t)((dw_id_l>>21) & 0xFF);
        pCAN_Channel->pCAN->txbuf.id.b[2] = (uint8_t)(((dw_id_l>>13) & 0xE0) | ((dw_id_l>>15) & 0x07));
        pCAN_Channel->pCAN->txbuf.id.b[1] = (uint8_t)((dw_id_l>>7) & 0xFF);
        pCAN_Channel->pCAN->txbuf.id.b[0] = (uint8_t)((dw_id_l<<1) & 0xFF);
        pCAN_Channel->pCAN->txbuf.id.b[1] |= MSCAN_SRR_MASK | MSCAN_IDE_MASK;          /* SRR=1, IDE=1 */
    }
    else
    {                                                       /* send standard id */
        pCAN_Channel->pCAN->txbuf.id.b[0] = (uint8_t)((dw_id_l>>3) & 0xFF);
        pCAN_Channel->pCAN->txbuf.id.b[1] = (uint8_t)((dw_id_l & 0x07)<<5);
    }

    pCAN_Channel->pCAN->txbuf.dlr.Byte = dlc;

    pCAN_Channel->pCAN->txbuf.dsr.w[0] = *(uint16_t*)&data[0];  /* copy data to the tx buffer */
    pCAN_Channel->pCAN->txbuf.dsr.w[1] = *(uint16_t*)&data[2];
    pCAN_Channel->pCAN->txbuf.dsr.w[2] = *(uint16_t*)&data[4];
    pCAN_Channel->pCAN->txbuf.dsr.w[3] = *(uint16_t*)&data[6];
                                                            /* send frame */
    pCAN_Channel->pCAN->cantflg.Byte = pCAN_Channel->pCAN->cantbsel.Byte;

    return (TRUE);                                          /* message is sent */
}

/****************************************************************************************************/
/**
 * @brief  CAN Application transmit message
 * @par    Description:
 *         provides a simple interface to the appliation for data transmission
 * @param  uint8_t   CAN_Channel     - numeric value of the selected hardware channel
 * @param  uint32_t  CAN_Id          - CAN identifier
 * @param  uint8_t   CAN_Datalength  - data length
 * @param  uint8_t * CAN_Data        - reference to data
 * @return TRUE - message send, FALSE - message not sent or channel not matching
 */
bool_t CAN_ApplicationTransmitMsg(uint8_t CAN_Channel, uint32_t CAN_Id, uint8_t CAN_Datalength, uint8_t * CAN_Data)
{
    CoreCANstruct *pCAN_Channel;
    bool_t retvalue;

    pCAN_Channel = &CAN_GetChannelAddress(CAN_Channel);

    if (pCAN_Channel == NULL)                               /* check if matching channel found */
    {
        retvalue = FALSE;                                   /* channel not active */
    }
    else
    {                                                       /* transmit message immediatley */
        retvalue = CAN_TransmitMsg(pCAN_Channel, CAN_Id, CAN_Datalength, CAN_Data);
    }
    return (retvalue);
}

/****************************************************************************************************/
/**
 * @brief  CAN get address of selected channel
 * @par    Description:
 *         provides a simple interface to the appliation for data transmission
 * @param  uint8_t   CAN_Channel     - numeric value of the selected hardware channel
 * @return CoreCANstruct address of channel
 */
void CAN_GetChannelAddress(uint8_t channel, CoreCANstruct * address)
{
    CoreCANstruct *pCAN_Channel;

    switch(CAN_Channel)                                     /* select CAN struct */
    {
        #ifdef USE_CAN0
        case 0:
            pCAN_Channel = &CoreTXCAN0;
            break;
        #endif
        default:
            pCAN_Channel = NULL;                            /* channel not active - signal this to */
            break;                                          /* calling function */
    }
    address = pCAN_Channel;
}

/****************************************************************************************************/
/* Interrupt section                                                                                */
/****************************************************************************************************/

/****************************************************************************************************/
/**
 * @brief  CAN receive and transmit interrupt functions
 * @par    Description:
 *         Call the tranmit/receive functions with the specific channel address pointer.
 * @return void
 */
#ifdef USE_CAN0
    INTERRUPT void CAN_Receive0_ISR(void)
    {
        (void)CAN_ReceiveFct(&CoreRXCAN0);                  /* call the receive function with the specific CAN struct */
    }

    INTERRUPT void CAN_Transmit0_ISR(void)
    {
        (void)CAN_TransmitFct(&CoreTXCAN0);
    }
#endif
